/*
 * d_gettype.c
 *
 * $Source: /nfs/source/MASTER/garden/datlib/RCS/d_gettype.c,v $
 * $Date: 94/01/07 21:59:47 $
 * $Author: sbeck $
 * $Locker:  $
 *
 * This file is part of the dat library.
 *
 * $Log:	d_gettype.c,v $
 * Revision 1.8  94/01/07  21:59:47  sbeck
 * Removed all evidence of floating point representation
 * since it is now completely handled by proper byte order
 * processing. Also changed usage of the predefined macro,
 * MACH_BYTE_ORDER to the function call mach_get_byte_order() 
 * which automatically determines the endian-ness of the host
 * machine.
 * 
 * Added handling of 'x' and 'X' datatypes.
 * 
 * Revision 1.7  1992/12/25  01:05:50  holtzman
 * updated for const
 *
 * Revision 1.6  92/04/09  17:02:45  holtzman
 * Routine checkin
 * 
 * Revision 1.4.3.1  92/04/09  16:54:05  holtzman
 * changed so that float and byte mismatches don't cause program to exit
 * 
 * Revision 1.4.1.1  91/11/02  23:31:24  klee
 * updated for const declarations
 * 
 * Revision 1.4  90/12/08  03:47:33  kaboom
 * Replaced under RCS, moved copyright message to copyright.h.
 * 
 * Revision 1.3  89/01/13  02:43:55  jwang
 * Add M.I.T. copyright message.
 * 
 * Revision 1.2  85/06/29  13:37:07  musicus
 * Changed type codes for integers and unsigned:
 * 	c,d,D = SIGNED1, SIGNED2, SIGNED4
 * 	b,u,U = UNSIGNED1, UNSIGNED2, UNSIGNED4
 * Also added d_settype() subroutine
 * Also modified datsubs.vr accordingly
 *
 * Revision 1.1  85/06/29  00:11:22  musicus
 * Initial revision
 */

static char rcsid[] = "$Header: /nfs/source/MASTER/garden/datlib/RCS/d_gettype.c,v 1.8 94/01/07 21:59:47 sbeck Exp $" ;

#include "copyright.h"
#include "dat.h"

/* Global variables - must reflect type definitions in dat.h */

static struct typ
{
    char 	*value;
    int 	 code;
    int 	 nbytes;
};

static struct typ stype[] = {
	{ASCII,			'a',		sizeof(char)},
	{VARIANT,		'v',		1},
	{INTEGER1,		'c',		1},
	{INTEGER2,		'd',		2},
	{INTEGER4,		'D',		4},
	{UNSIGNED1,		'b',		1},
	{UNSIGNED2,		'u',		2},
	{UNSIGNED4,		'U',		4},
	{FLOAT,			'f',		sizeof(float)},
	{DOUBLE,		'F',		sizeof(double)},
	{COMPLEX,		'z',		2*sizeof(float)},
	{DCOMPLEX,		'Z',		2*sizeof(double)},
	{FIX2,			'x',		2},
	{FIX4,			'X',		4},
	{(char*)NULL,		 0,		0}
} ;


/*
 * int d_gettype(d,&nbps)
 * 
 *    Reads value of the "type" keyword from the DATFILE d, looks it up
 *    in a table of known types and returns the appropriate code.  If
 *    type is unknown or type keyword is missing, then returns 0.
 *    If &nbps argument is non-zero, sets the integer variable nbps to
 *    the number of bytes per sample for the given data type.
 * 
 *    Inputs:
 *        d - DATFILE pointer
 *    Outputs:
 *        nbps - address of integer variable where will store the number of
 *               bytes per sample for this data type
 *    Return value:
 *        code for the type of the dat file (see below)
 *        0 if type not found (sets *npbs to 1)
 */

int 
d_gettype(d, nbps)
const DATFILE	*d;
int 	*nbps;
{
    register	int 	 i;
    const char 		*p;

    if ((p = d_getkey(d, TYPE)))
      {
	for(i = 0; stype[i].value; i++)
	  {
	    if(!strcmp(p, stype[i].value))
	      {
		if(nbps) 
		  {
		    *nbps = stype[i].nbytes;
		  }
		
		return(stype[i].code);
	      }
	  }
      }

    if(nbps) 
    {
	*nbps = 1;
    }

    return(0);
}


/*
 *
 * int d_settype(d,type)
 *
 *    Looks up the single character code "type" in the table below, then
 *    sets the TYPE keyword in DATFILE "d" to the corresponding string value.
 *
 *    Inputs:
 *        d - DATFILE pointer
 *        type - single character type code
 *    Outputs:
 *        d_setkey() on TYPE keyword
 *    Return value:
 *        0 if finds type code in table
 *        -1 if not find type code in table
 */

int 
d_settype(d, type)
DATFILE	*d;
int 	 type;
{
    register	int 	i;
    enum byte_order	bo ;

    for(i = 0; stype[i].value; i++)
    {
	if(stype[i].code == type)
	{
	    d_setkey(d, TYPE, stype[i].value);

	    /* change the arch flags */
	    d_flush(d) ;

	    /* what checks are made depends on the file type */
	    switch (type) {
	    case 'f':       /* float  */
	    case 'F':       /* double */
	    case 'z':       /* complex float  */
	    case 'Z':       /* complex double */
	    case 'd':       /* 2 byte signed integer */
	    case 'D':       /* 4 byte signed integer */
	    case 'u':       /* 2 byte unsigned integer */
	    case 'U':       /* 4 byte unsigned integer */
	    case 'x':	    /* 2 byte fixed point */
	    case 'X':	    /* 4 byte fixed point */
	    case NOKEY:     /* unsure of type, check bo */
	      bo = d_get_byte_order(d) ;
	      if (bo == NOKEY) {
		/* byte order flag non-existent */
		d_set_byte_order(d, mach_get_byte_order()) ;
	      }
	      else {
		/* byte order flag exists */
		if (mach_get_byte_order() != bo) {
		  fprintf(stderr,"ATTN: byte order conversion will occur.\n");
		  /* fprintf(stderr, "Exiting...\n") ; */
		  /* exit (-1) ; */
		}
	      }
	      break ;
	      
	      /* for the following cases, don't need
		 to do any architecture checks */
	    case 'a':       /* 1 byte ascii */
	    case 'v':       /* 1 byte variant */
	    case 'c':       /* 1 byte signed integer */
	    case 'b':       /* 1 byte unsigned integer */
	    default:        /* a type no ones heard of */
	      break ;
	    }
	    
	    d_flush(d) ;

	    
	    return(0);
	}
    }

    return(-1);
}







