// -------------------------------------------------------------------
// tkTimeline.cc
//
// tk widget to display a timeline label
// dpwe@mit.edu 1994sep14
//

#include <assert.h>
#include <math.h>

extern "C" {
#include "tcl.h"
#include "tkConfig.h"
#include "tk.h"
}

// Widget data structure

typedef struct {
    // standard fields
    Tk_Window	tkwin;
    Display	*display;
    Tcl_Interp	*interp;
    Tk_3DBorder bgBorder;	/* Used for drawing background. */
    XColor	*fgColor;
    int 	borderWidth;
    int 	padX;
    int 	padY;
    int 	width, height;
    int 	relief;		// appearance of whole window
    // data for the drawing activities
    GC  	gc;
    XFontStruct	*font;
    int 	doubleBuffer;	// do we double buffer?
    int 	updatePending;
    // fields describing the content
    int 	isLog;
    int		isVertical;
    int		tixTopLeft;
    int 	sense;
    int 	isSeconds;	// treat the values as seconds, i.e. make mins
    int		tickLen;
    double	minVal;
    double	maxVal;
} Timeline;

extern "C" {
// static function prototypes

int Timeline_Init(Tcl_Interp *interp);
    /* add the tcl commands defined in this package to the interpreter */
int TimelineCmd(ClientData clientData, Tcl_Interp *interp, 
	       int argc, char **argv);
static int  TimelineConfigure(Tcl_Interp *interp, Timeline *sw, 
			     int argc, char **argv, int flags);
static void TimelineDestroy(ClientData clientData);
static void TimelineDisplay(ClientData clientData);
static void TimelineEventProc(ClientData clientData, XEvent *eventPtr);
static int  TimelineWidgetCmd(ClientData clientData, Tcl_Interp * interp, 
			     int argc, char **argv);
};

// Argv parsing info

static Tk_ConfigSpec timelineConfigSpecs[] = {
    {TK_CONFIG_BORDER, "-background", "background", "Background", 
     "#cdb79e", Tk_Offset(Timeline, bgBorder), 0 }, 
    {TK_CONFIG_SYNONYM, "-bg", "background", NULL, NULL, 0, 0}, 
    {TK_CONFIG_COLOR, "-foreground", "foreground", "Foreground", 
     "black", Tk_Offset(Timeline, fgColor), 0 }, 
    {TK_CONFIG_SYNONYM, "-fg", "foreground", NULL, NULL, 0, 0}, 
    {TK_CONFIG_PIXELS, "-borderwidth", "borderWidth", "BorderWidth",
	"2", Tk_Offset(Timeline, borderWidth), 0},
    {TK_CONFIG_SYNONYM, "-bw", "borderWidth", NULL, NULL, 0, 0}, 
    {TK_CONFIG_PIXELS, "-padX", "padX", "PadX",
	"2", Tk_Offset(Timeline, padX), 0},
    {TK_CONFIG_PIXELS, "-padY", "padY", "PadY",
	"2", Tk_Offset(Timeline, padY), 0},
    {TK_CONFIG_PIXELS, "-width", "width", "Width",
	"400", Tk_Offset(Timeline, width), 0},
    {TK_CONFIG_PIXELS, "-height", "height", "Height",
	"32", Tk_Offset(Timeline, height), 0},
    {TK_CONFIG_RELIEF, "-relief", "relief", "Relief",
	"raised", Tk_Offset(Timeline, relief), 0},
    {TK_CONFIG_INT, "-dbl", "doubleBuffer", "DoubleBuffer",
	"1", Tk_Offset(Timeline, doubleBuffer), 0},
    {TK_CONFIG_FONT, "-font", "font", "Font",
	"fixed", Tk_Offset(Timeline, font), 0},
    {TK_CONFIG_BOOLEAN, "-isLog", "isLog", "IsLog", 
        "0", Tk_Offset(Timeline, isLog), 0}, 
    {TK_CONFIG_BOOLEAN, "-isVertical", "isVertical", "IsVertical", 
        "0", Tk_Offset(Timeline, isVertical), 0}, 
    {TK_CONFIG_BOOLEAN, "-isSeconds", "isSeconds", "IsSeconds", 
        "0", Tk_Offset(Timeline, isSeconds), 0}, 
    {TK_CONFIG_BOOLEAN, "-tixTopLeft", "tixTopLeft", "TixTopLeft", 
        "0", Tk_Offset(Timeline, tixTopLeft), 0}, 
    {TK_CONFIG_BOOLEAN, "-sense", "sense", "Sense", 
        "1", Tk_Offset(Timeline, sense), 0}, 
    {TK_CONFIG_PIXELS, "-tickLen", "tickLen", "TickLen",
	"2", Tk_Offset(Timeline, tickLen), 0},
    {TK_CONFIG_DOUBLE, "-minVal", "minVal", "MinVal",
	"0.0", Tk_Offset(Timeline, minVal), 0},
    {TK_CONFIG_DOUBLE, "-maxVal", "maxVal", "MaxVal",
	"1.0", Tk_Offset(Timeline, maxVal), 0},
    {TK_CONFIG_END, NULL, NULL, NULL, NULL, 0, 0}
};

int Timeline_Init(Tcl_Interp *interp)
{   /* add the tcl commands defined in this package to the interpreter */
    Tk_Window main = Tk_MainWindow(interp);

    Tcl_CreateCommand(interp, "timeline", TimelineCmd, (ClientData) main,
	    (void (*)(ClientData)) NULL);
    return TCL_OK;
}

int TimelineCmd(ClientData clientData, Tcl_Interp *interp, 
	       int argc, char **argv)
{   /* Tcl command to instance a new Timeline widget */
    Tk_Window	main = (Tk_Window) clientData; 
    Timeline 	*tmln; 
    Tk_Window	tkwin;

    if(argc < 2)  {
	Tcl_AppendResult(interp, "wrong # args: should be\"", 
			 argv[0], " pathName ?options?\"", NULL);
	return TCL_ERROR;
    }
    tkwin = Tk_CreateWindowFromPath(interp, main, argv[1], NULL);
    if(tkwin == NULL)  {
	return TCL_ERROR;
    }
    Tk_SetClass(tkwin, "Timeline");

    // Allocate and initialize the structure
    tmln = (Timeline *)ckalloc(sizeof(Timeline));
    tmln->tkwin   = tkwin;
    tmln->display = Tk_Display(tkwin);
    tmln->interp  = interp;
    tmln->bgBorder= NULL;
    tmln->fgColor = NULL;
    tmln->gc      = None;
    tmln->font	  = None;
    tmln->doubleBuffer = 1;
    tmln->updatePending = 0;

    Tk_CreateEventHandler(tmln->tkwin, ExposureMask|StructureNotifyMask,
			  TimelineEventProc, (ClientData)tmln);
    Tcl_CreateCommand(interp, Tk_PathName(tmln->tkwin), TimelineWidgetCmd, 
		      (ClientData)tmln, (void (*)(ClientData))NULL);
    if(TimelineConfigure(interp, tmln, argc-2, argv+2, 0) != TCL_OK)  {
	Tk_DestroyWindow(tmln->tkwin);
	return TCL_ERROR;
    }
    interp->result = Tk_PathName(tmln->tkwin);
    return TCL_OK;
}

static void TimelineGetWorkArea(Timeline *tmln, 
				int *px, int *py, int *pw, int *ph)
{   /* return the area that time line legend is actually drawn in */
    Tk_Window tkwin = tmln->tkwin;
    int x, y, w, h;

    *px = tmln->padX;
    *pw = Tk_Width(tkwin) - 2*tmln->padX;
    *py = tmln->padY;
    *ph = Tk_Height(tkwin) - 2*tmln->padY;
}

static int TimelineWidgetCmd(ClientData clientData, Tcl_Interp *interp, 
			    int argc, char **argv)
{   /* passed commands executed on the widget */
    Timeline	*tmln = (Timeline *)clientData;
    int 	result = TCL_OK;
    int 	length;
    char	c;

    if(argc < 2)  {
	Tcl_AppendResult(interp, "wrong # args: should be \"", 
			 argv[0], " option ?arg arg ...?\"", NULL);
	return TCL_ERROR;
    }
    Tk_Preserve((ClientData)tmln);
    c = argv[1][0];
    length = strlen(argv[1]);
    if( (c=='c') && (strncmp(argv[1], "configure", length)==0) )  {
	if(argc==2) {
	    result =Tk_ConfigureInfo(interp, tmln->tkwin, timelineConfigSpecs, 
				      (char *)tmln, NULL, 0);
	} else if(argc==3) {
	    result =Tk_ConfigureInfo(interp, tmln->tkwin, timelineConfigSpecs, 
				      (char *)tmln, argv[2], 0);
	} else {
	    result =TimelineConfigure(interp, tmln, argc-2, argv+2, 
				      TK_CONFIG_ARGV_ONLY);
	}
    } else if( (c == 'v') && (strncmp(argv[1], "valOfPixel", length)==0) )  {
	if(argc != 3)  {
	    Tcl_AppendResult(interp, "wrong # args: should be \"", 
			     argv[0], " valOfPixel pixel\"", NULL);
	    return TCL_ERROR;
	}
	// Convert the passed pixel into a value
	int pixel, base, width, x,y,w,h;
	double val;
	TimelineGetWorkArea(tmln, &x, &y, &w, &h);
	if(tmln->isVertical)  {
	    base = y+h; width = -h;
	} else {
	    base = x; width = w;
	}
	if(Tcl_GetInt(interp, argv[2], &pixel)!=TCL_OK) {
	    return TCL_ERROR;
	}
	if(tmln->isLog) {
	    val = tmln->minVal * exp(log(tmln->maxVal/tmln->minVal)
				     *double(pixel-base)/double(width));
	} else {
	    val = tmln->minVal + (tmln->maxVal-tmln->minVal)
				     *double(pixel-base)/double(width);
	}
	sprintf(interp->result, "%f", val);
    } else if( (c == 'p') && (strncmp(argv[1], "pixelOfVal", length)==0) )  {
	if(argc != 3)  {
	    Tcl_AppendResult(interp, "wrong # args: should be \"", 
			     argv[0], " pixelOfVal value\"", NULL);
	    return TCL_ERROR;
	}
	// Convert the passed value into a pixel
	int pixel, base, width, x,y,w,h;
	double val;
	TimelineGetWorkArea(tmln, &x, &y, &w, &h);
	if(tmln->isVertical)  {
	    base = y+h; width = -h;
	} else {
	    base = x; width = w;
	}
	if(Tcl_GetDouble(interp, argv[2], &val)!=TCL_OK) {
	    return TCL_ERROR;
	}
	if(tmln->isLog)  {
	    pixel = base + (int)rint(double(width) *
		 log(val/tmln->minVal)/log(tmln->maxVal/tmln->minVal));
	} else {
	    pixel = base + (int)rint(double(width) *
		 (val - tmln->minVal)/(tmln->maxVal-tmln->minVal));
	}
	sprintf(interp->result, "%d", pixel);
    } else {
	Tcl_AppendResult(interp, "bad option \"", argv[1],
			 "\":  must be configure, valOfPixel or pixelOfVal", 
			 (char *) NULL);
	result = TCL_ERROR;
    }
    if (result == TCL_OK && !tmln->updatePending) {
	Tk_DoWhenIdle(TimelineDisplay, (ClientData) tmln);
	tmln->updatePending = 1;
    }
    Tk_Release((ClientData)tmln);
    return result;
}

static int TimelineConfigure(Tcl_Interp *interp, Timeline *tmln, 
			    int argc, char **argv, int flags)
{   /* handle configure commands */
    int rslt = TCL_OK;

    if(Tk_ConfigureWidget(interp, tmln->tkwin, timelineConfigSpecs, 
			  argc, argv, (char *)tmln, flags)!= TCL_OK)  {
	return TCL_ERROR;
    }
    /* ensure values are OK */
    if(tmln->isLog)  {
	if( ! ((tmln->minVal * tmln->maxVal) > 0) )  {
	    fprintf(stderr, "TimelineConfigure: for log line, both args must be nonzero and of same sign\nnot like %f and %f\n", tmln->minVal, tmln->maxVal);
	    tmln->minVal = 0.1; 	tmln->maxVal = 1.0;
	}
    }
    // set the colors and GC for the redisplay
    Tk_SetWindowBackground(tmln->tkwin, Tk_3DBorderColor(tmln->bgBorder)->pixel);
    if (tmln->gc == None) {
	XGCValues gcValues;
	gcValues.function = GXcopy;
	gcValues.graphics_exposures = False;
	tmln->gc = Tk_GetGC(tmln->tkwin,GCFunction|GCGraphicsExposures, &gcValues);
    }
    XSetFont(Tk_Display(tmln->tkwin), tmln->gc, tmln->font->fid);
    // register desired window geometry, arrange for redisplay
    Tk_GeometryRequest(tmln->tkwin, tmln->width, tmln->height);
    Tk_SetInternalBorder(tmln->tkwin, tmln->borderWidth);
    if (!tmln->updatePending) {
	Tk_DoWhenIdle(TimelineDisplay, (ClientData)tmln);
	tmln->updatePending = 1;
    }
    return rslt;
}

static void TimelineEventProc(ClientData clientData, XEvent *eventPtr)
{				/* handle events for the Timeline widget */
    Timeline *tmln = (Timeline *)clientData;
    
    if(eventPtr->type == Expose)  {
	if(!tmln->updatePending)  {
	    Tk_DoWhenIdle(TimelineDisplay, (ClientData)tmln);
	    tmln->updatePending = 1;
	}
    } else if(eventPtr->type == ConfigureNotify)  {
	// adjust self for resize 
	if(!tmln->updatePending)  {
	    Tk_DoWhenIdle(TimelineDisplay, (ClientData)tmln);
	    tmln->updatePending = 1;
	}
    } else if(eventPtr->type == DestroyNotify)  {
	Tcl_DeleteCommand(tmln->interp, Tk_PathName(tmln->tkwin));
	tmln->tkwin = NULL;
	if(tmln->updatePending)  {
	    Tk_CancelIdleCall(TimelineDisplay, (ClientData)tmln);
	}
	Tk_EventuallyFree((ClientData)tmln, TimelineDestroy);
    }
}

static void drawX(Tk_Window tkwin, GC gc, Drawable d=None, int inset=0)
{				/* write an X into a tkwin */
    Display *disp = Tk_Display(tkwin);
    int x, y, w, h;
    
    if(d==None)
	d = Tk_WindowId(tkwin);
    
    x = inset;
    w = Tk_Width(tkwin) - 2*x;
    y = inset;
    h = Tk_Height(tkwin) - 2*y;
    
    XDrawLine(disp, d, gc, x, y, x+w, y+h);
    XDrawLine(disp, d, gc, x, y+h, x+w, y);
}

#define MAX(a,b) (((a)>(b))?(a):(b))

/************************* TIMELINE CODE from ~/lib/wndw.c 1994sep21 *****/

static void NmToString(int n, char *s)
    {	/* was itoa from K&R via sndf.c */
    int     i,j;
    long    sign;
    char    c;

    if( (sign = n) < 0)     n = -n;
    i = 0;
    do	{
	s[i++] = n%10 + '0';
	} while( (n /= 10) > 0);
    if(sign < 0)
	s[i++] = '-';
    s[i--] = '\0';
    for(j=0; j<i; ++j,--i)	/* reverse the string */
	{ c = s[i]; s[i] = s[j]; s[j] = c; }
    }

static void FltToString(double f, char *s, int dp)
	{	/* convert a float into an ascii string,  given # of DP */
	long	l;
	char	t[32];
	double	k = 1.0;
	int 	i;
	double	rup = 0.5;		/* roundup */

	for(i = dp; i>0; --i, rup /= 10.0)
		;	        /* form rup as 1/2 smallest digit (per DP) */
	if(f > 0)
	    f += rup;		/* Add on, so truncation gives right result */
	else 			/* (shift other way to round -ve nums) */
	    f -= rup;
	l = (long)f;
	NmToString(l,s);		/* convert int part */
	if(dp > 0)
		{
		for(i=0; i++ < dp; k *= 10 )	/* K scales for dp */
			;
		l = (long)(k*(1.0+f - (float)l));
		NmToString(l,t);
		t[0] = '.';	      /* overwrite place holding 1, add dp */
		strcat(s,t);
		}
	}

static void WWriteAt(Tk_Window tkwin, Drawable d, GC gc, 
		     int x, int y, char *str)
    {
static XTextItem txitem[1];
static int init = 0;
    int     nc, dir, fa, fd;
    XCharStruct	xcs;
    Display *disp = Tk_Display(tkwin);

    if(init==0)
	{
	txitem[0].delta  = 1;
	txitem[0].font   = None;
	init = 1;
	}
    nc = strlen(str);
    XQueryTextExtents(disp, XGContextFromGC(gc),
		      str, nc, &dir, &fa, &fd, &xcs);
    txitem[0].chars  = str;
    txitem[0].nchars = nc; 
    XDrawText(disp, d, gc, x, y+fa, txitem, 1);
    }

static int WTextWidth(Tk_Window tkwin, GC gc, char *str)
{
    XCharStruct xcs;
    int         nc, dir, fa, fd;
    Display *disp = Tk_Display(tkwin);

    nc = strlen(str);
    XQueryTextExtents(disp, XGContextFromGC(gc),
		      str, nc, &dir, &fa, &fd, &xcs);
    return xcs.width;
}

static int WTextHeight(Tk_Window tkwin, GC gc, char *str)
{
    XCharStruct xcs;
    int         nc, dir, fa, fd;
    Display *disp = Tk_Display(tkwin);

    nc = strlen(str);
    XQueryTextExtents(disp, XGContextFromGC(gc),
		      str, nc, &dir, &fa, &fd, &xcs);
    return fa+fd;
}

#define LALIGN -1
#define CALIGN 0
#define RALIGN 1

static void WWriteIn(Tk_Window tkwin, Drawable d, GC gc, 
		     XRectangle *pr, 
		     char *str, int lcrAlign)	/* write string in rectangle */
    {
    int     strwidth;
    int     strheight;
    int     acty;

    strwidth = WTextWidth(tkwin, gc, str);
    strheight = WTextHeight(tkwin, gc, str);
    /* always centrally aligned in vert */
    acty = pr->y+(pr->height - strheight)/2;	/* acty is top of text */
    if(lcrAlign == LALIGN)		/* left align */
	WWriteAt(tkwin, d, gc, pr->x, acty, str);
    else if(lcrAlign == RALIGN)	/* right align */
	WWriteAt(tkwin, d, gc, pr->x + pr->width - strwidth, acty, str);
    else			/* CALIGN i.e. centred */
	WWriteAt(tkwin, d, gc, pr->x + (pr->width - strwidth)/2, acty, str);
    }

static float WNearestRoundNum(double num, double *sub)
    {	/* return next size down too as sub-step */
static int	possVals = 3;
static double	rndNums[3][2] = {{ .707, 1.0 },{ 1.414, 2.0 },{ 3.163, 5.0}};
static double	oom = 10.0;	/* round number configuration */
    double	scale = 1.0;
    double	ref;
    int 	i;

    ref = rndNums[0][0];	/* lower bound for lowest round number */
    if(num < 0)		{ num *= -1.0; 	scale *= -1.0; }
    while(num < ref)
	{
	num *= oom;	/* next order of magnitude */
	scale /= oom;	/* keep track of order of magnitude */
	}
    while(num >= oom*ref)
	{
	num /= oom;	/* next order of magnitude */
	scale *= oom;	/* keep track of order of magnitude */
	}
    /* new num is in range ref <= num < oom*ref */
    /* quantize.. */
    i = possVals - 1;
    while(num < rndNums[i][0] && i>0)
	--i;
    num = scale * rndNums[i][1];
    if(i < 2)
	*sub = scale * rndNums[i-2+possVals][1] / oom;
    else
	*sub = scale * rndNums[i-2][1];
    return num;
    }

static void WReplaceOrAppend(char *s, char c1, char c2)
{   /* target string, char to find, char to replace it with- or stick on end */
    char *t;

    if( (t = strchr(s, c1)) == NULL)
	{ t = s + strlen(s);	*(t+1) = '\0';	} /* hope for space */
    *t = c2;
    }

static int WDecPoints(double step) 
{   /* return num of decimal points to code a step */
    double ref = 1.0;
    int   pts = 0;

    if(step<0) step = -step;
    if(step==0)	return 0;
    while(ref>step)
	{ ref /= 10.0; ++pts;  }
    return pts;
    }

#define     TICKSPC     50	/* ideal # pixels between ticks */
#define     HTICKSPC    70	/* ideal # pixels between horiz ticks */
#define     VTICKSPC    20	/* ideal # pixels between vert ticks */
#define     VOFFS       3	/* Border to offset text from ticks */
#define     HOFFS       5	/* Border to offset text from ticks */

static void WAddTimeLine(Tk_Window tkwin, Drawable d, GC gc, 
			 XRectangle *pr, 
			 double t1, double t2,
			 int vnh, int sns, int und, 
			 int tickLen, int tix[], int *pntx) 
{   /* write linear scale into a box */
    double  dt;
    short   dx,xb,xi,tkh,guardx,tkspc;
    double  xPerT, tPerTk, xPerTk;
    int     tkPerIx, tkix;
    long    n0;
    double  nrst, pwr, subdiv;
    double  t0, x0;
    int     dp,nc;
    char    s[32];
    XRectangle    trec;
    double  rescale = 1.0;
    char    rescCh = '.';
    int     cwidth = WTextWidth(tkwin, gc, " ");
    int     lineh  = WTextHeight(tkwin, gc, "H");
    Display *disp = Tk_Display(tkwin);

    trec  = *pr;
    if(pntx) *pntx = 0;

    dt = t2-t1;
    if(dt==0.0) return;
    if(dt<0)	dt = -dt;   /* args in wrong order ? */
    /* choose SI range for units & labels */
    if(dt>5000)                       { rescale = 0.001;      rescCh = 'k'; }
    else if(dt>5000000)               { rescale = 0.000001;   rescCh = 'M'; }
    else if(dt<0.5 && t2 <1.0)        { rescale = 1000.0;     rescCh = 'm'; }
    else if(dt<0.0005 && t2 <0.001)   { rescale = 1000000.0;  rescCh = 'u'; }
    
    if(vnh)
	{
	dx = pr->height;
	xb = pr->y;
	trec.height = 0;
	trec.width -= cwidth;
	trec.x += cwidth/2;
	tkspc = (short)(1.5*lineh);
	guardx = tkspc/2;
	}
    else{
	dx = pr->width;
	xb = pr->x;
	trec.width = 0;
	trec.height -= 2*VOFFS;
	trec.y += VOFFS + (und?0:-(tickLen-1));
	tkspc = WTextWidth(tkwin, gc, "0.0000   ");
	guardx = WTextWidth(tkwin, gc, "0.0000");
	}
    xPerT = ((float)dx)/dt;
    tPerTk  = ((float)tkspc)/xPerT;	/* ideal time spacing */
    /* find 1, 2 or 5 x 10^n closest to tPerTk */
    tPerTk = WNearestRoundNum(tPerTk, &subdiv);
    tkPerIx = (int)(.5+tPerTk/subdiv);	/* had a rounding problem here */
    dp = WDecPoints(tPerTk*rescale);
    tPerTk = subdiv;	/* actually refer to period of ticks */
    xPerTk = tPerTk * xPerT;
    n0 = (long)(.99+(t1/tPerTk)); 
    t0 = tPerTk*(float)n0;         /* earliest time exact multiple of tPerTk */
    tkix = n0%tkPerIx;

    x0 = (t0-t1)*xPerT;
    while(x0 < dx)
	{
	if(sns)
	    xi = xb+(short)x0;
	else
	    xi = xb + dx - (short)x0;
	tkh = tickLen;
	if(tkix==0 && x0<dx-guardx)
	    {
	    FltToString(t0*rescale,s,dp);
	    if(rescale != 1.0)
		WReplaceOrAppend(s, '.', rescCh); /* replc dp with e.g. 'k' */
	    if(vnh)
		{  trec.y = xi-lineh/2; 
		   WWriteIn(tkwin, d, gc, &trec, s, RALIGN);	}
	    else
		{  trec.x = xi;
		   WWriteIn(tkwin, d, gc, &trec, s, LALIGN); }

	    tkh = 2*tickLen;
	    if(tix)  *tix++ = xi;	/* remember where we put these ticks */
	    if(pntx) ++(*pntx);
	    }
	if(vnh)
	    {  XDrawLine(disp, d, gc, und?pr->x:pr->x+pr->width, xi,
			 und?(pr->x+tkh):(pr->x+pr->width-tkh), xi);  }
	else
	    {  XDrawLine(disp, d, gc, xi, und?pr->y:pr->y+pr->height, 
			 xi, und?(pr->y+tkh):(pr->y+pr->height-tkh));  }
	t0 += tPerTk;
	x0 += xPerTk;
	tkix = (tkix +1 )%tkPerIx;
	}
    }

void WAddLogTimeLine(Tk_Window tkwin, Drawable d, GC gc, XRectangle *pr,
		     double t1, double t2, int vnh, int sns, int und,
		     int tickLen, int tix[], int *pntx)
{   /* write logarithmic scale into a box */
    double  dt, pixScale, idealStep;
    short   dx,xb,xi,tkh;
    double  t0, nextt0, nt0, nextdt, x0, tkdt, nexttkdt;
    int     dp,nc,tkspc;
    char    s[32];
    int     cwidth = WTextWidth(tkwin, gc, " ");
    int     lineh  = WTextHeight(tkwin, gc, "H");
    XRectangle   trec;
    Display *disp = Tk_Display(tkwin);

    trec = *pr;
    if(pntx) *pntx = 0;

    if(t1*t2 <= 0.0)	{ fprintf(stderr,"LogLine: invalid %f..%f\n", t1,t2);
			  abort(); }
    dt = t2-t1;
    if(dt==0.0) return;
/*    if(dt<0)	dt = -dt;   /* args in wrong order ? */
    if(vnh)
	{
	dx = pr->height - 1;
	xb = pr->y+1;
	trec.height = 0;
	trec.x += cwidth/2;	trec.width -= cwidth;
	tkspc = (int)(1.5*lineh);
	}
    else{
	dx = pr->width - 1;
	xb = pr->x + 1;
	trec.width = 0;
	trec.y += VOFFS + (und?0:-(tickLen-1));
	trec.height -= 2*VOFFS + (und?0:-(tickLen-1));
	tkspc = WTextWidth(tkwin, gc, "0.0000   ");
	}
    pixScale = log(t2/t1)/dx;	/* so t(x) = t1 * exp(pixscale*(x-left)) */
    idealStep = exp(pixScale*tkspc);
    if(idealStep>2.0)	idealStep = 2.0;	/* else lose orders of mag */
    /* 2.0 is the smallest ratio between adjacent 'RoundNumbers' - I know */
/*    fprintf(stderr,"idealStep %f\n",idealStep);	*/

    dt = WNearestRoundNum(t1 * (idealStep-1.0), &tkdt);
    t0 = dt * MAX(1, (int)(.0 + t1/dt));	/* hacked- blocking 1993apr */
    dt = WNearestRoundNum(t0 * (idealStep-1.0), &tkdt);
    t0 = dt * (int)(.5 + t0/dt);

    while( (x0 = log(t0/t1)/pixScale) < dx)
	{
/*	fprintf(stderr, "t0 %.0f dt %.0f tkdt %.0f\n", t0, dt, tkdt);	/* */
/*	x0 = log(t0/t1)/pixScale; */
	if(x0>=0)
	    {
	    if(sns)
		xi = xb+(short)x0;
	    else
		xi = xb + dx - (short)x0;
	    dp = WDecPoints(dt);
	    FltToString(t0,s,dp);
	    if(vnh)
		{
		trec.y = xi /* - lineh/2 */;  /* dpwe 1993mar14: need lower */
		WWriteIn(tkwin, d, gc, &trec, s, RALIGN);
		}
	    else{
		trec.x = xi;
		WWriteIn(tkwin, d, gc, &trec, s, LALIGN);
		}

	    tkh = 2*tickLen;
	    if(tix)   *tix++ = xi;	/* remember where we put these ticks */
	    if(pntx)  ++(*pntx);
	    if(vnh)
		{
		XDrawLine(disp, d, gc, und?pr->x:pr->x+pr->width, xi, 
			  und?(pr->x+tkh):(pr->x+pr->width-tkh), xi);
		}
	    else
		{
		XDrawLine(disp, d, gc, xi, und?pr->y:pr->y+pr->height, 
			  xi, und?(pr->y+tkh):(pr->y+pr->height-tkh));
		}
	    }
	nextt0 = t0+dt;
	nextdt = WNearestRoundNum(nextt0 * (idealStep-1.0), &nexttkdt);
	nextt0 = dt * (int)(.5 + nextt0/dt);
	nextdt = WNearestRoundNum(nextt0 * (idealStep-1.0), &nexttkdt);
	nt0    = nextdt * (int)(.51 + nextt0/nextdt);
	/* had to make it .51 to make .35 round to .4 when 350 made 400 */ 
	if(nt0 > nextt0)	nextt0 = nt0;	/* poke & hope for 1, 2, 5 */
	tkh = tickLen;
	while(((t0 += tkdt) < (.99*nextt0)) && (x0 = log(t0/t1)/pixScale)<dx )
	    if(x0>0) 
		{
		if(sns)
		    xi = xb+(short)x0;
		else
		    xi = xb + dx - (short)x0;
		if(vnh)
		    {
		    XDrawLine(disp, d, gc, und?pr->x:pr->x+pr->width, xi, 
			      und?(pr->x+tkh):(pr->x+pr->width-tkh), xi);
		    }
		else
		    {
		    XDrawLine(disp, d, gc, xi, und?pr->y:pr->y+pr->height, 
			      xi, und?(pr->y+tkh):(pr->y+pr->height-tkh));
		    }
		}
	t0 = nextt0;
	dt = nextdt;
	tkdt = nexttkdt;
	}
    }

static void TimelineDraw(Timeline *tmln, Drawable d=None)
{   /* actually draw the timeline data */
    Tk_Window 		tkwin = tmln->tkwin;
    GC 			gc    = tmln->gc;
    Display 		*disp = Tk_Display(tkwin);
    XColor 		*fgColor = tmln->fgColor;
    int x, y, w, h, y0, yscale, yl, yh;
    int start, end;
    int pts, i;
    double min, max;
    XRectangle rect;

    if(d==None)
	d = Tk_WindowId(tkwin);
    
    XSetForeground(disp, gc, tmln->fgColor->pixel);

    if(0)  {
	drawX(tkwin, gc, d, tmln->borderWidth+MAX(tmln->padX,tmln->padY));
	return;
    }

    TimelineGetWorkArea(tmln, &x, &y, &w, &h);
    rect.x = x; rect.y = y; rect.width = w; rect.height = h;
    if(tmln->isLog)  {
	WAddLogTimeLine(tkwin, d, gc, &rect, 
			tmln->minVal, tmln->maxVal, tmln->isVertical, 
			tmln->sense, tmln->tixTopLeft, 
			tmln->tickLen, NULL, NULL);
    } else {
	WAddTimeLine(tkwin, d, gc, &rect, 
		     tmln->minVal, tmln->maxVal, tmln->isVertical, 
		     tmln->sense, tmln->tixTopLeft, 
		     tmln->tickLen, NULL, NULL);
    }
}    

static void TimelineDisplay(ClientData clientData)
{				/* actually draw the contents of the widget */
    Timeline *tmln = (Timeline *)clientData;
    Tk_Window tkwin = tmln->tkwin;
    Pixmap pm = None;
    Drawable d;
    Display *disp = tmln->display;
    int 	tw, th, x, y, w, h, inset;
    
    tmln->updatePending = 0;
    if(!Tk_IsMapped(tkwin)) {
	return;
    }
    
    tw = Tk_Width(tkwin);
    th = Tk_Height(tkwin);
    
    // Create pixmap for double-buffering if necessary
    if(tmln->doubleBuffer)  {
	pm = XCreatePixmap(disp, Tk_WindowId(tkwin), tw, th, 
			   DefaultDepthOfScreen(Tk_Screen(tkwin)));
	d = pm;
    } else {			// draw directly into window
	d = Tk_WindowId(tkwin);
    }
    // redraw background
    Tk_Fill3DRectangle(disp, d, tmln->bgBorder, 0, 0, tw, th, tmln->borderWidth, 
		       tmln->relief);
    // Actually do the drawing
    inset = 2;
    
    //    drawX(tkwin, tmln->gc, d, tmln->borderWidth + inset);
    TimelineDraw(tmln, d);
    
    // copy to screen if double-buffering
    if(tmln->doubleBuffer)  {
	XCopyArea(disp, pm, Tk_WindowId(tkwin), tmln->gc, 0, 0, 
		  tw, th, 0, 0);
	XFreePixmap(disp, pm);
    }
}

static void TimelineDestroy(ClientData clientData)
{				/* destroy one of our objects */
    Timeline *tmln = (Timeline *)clientData;
    
    Tk_FreeOptions(timelineConfigSpecs, (char *)tmln, tmln->display, 0);
    if(tmln->gc != None)  {
	Tk_FreeGC(tmln->display, tmln->gc);
    }
    // don't destroy sound as we didn't new it
    ckfree((char *)tmln);
}


