/* 
 * tohex.c
 * convert a soundfile to a hex data block that image can read
 * dpwe@media.mit.edu 1994nov02
 */

/* std libs */
#include <stdlib.h>
#include <string.h>
/* personal libs */
#include <dpwelib.h>
#include <genutils.h>
#include <math.h>
#include <vecops.h>
#include <snd.h>

#ifdef SGI_FFT
#include "sgi_fft.h"
#else
#include <fft.h>
#endif /* SGI_FFT */

#ifdef DODISP
#include <window.h>
#endif /* DODISP */

#ifdef USEDAT
#include <dat/dat.h>
#include <time.h>	/* for datestamping dat files; not important */
#endif /* USEDAT */

/* constants */
#ifdef USEDAT
#define DFLT_EXTN	".dat"
#else /* !USEDAT */
#define DFLT_EXTN	".hex"	/* default extension for output */
#endif /* USEDAT */
#define DFLT_FFTO	1024
#define DFLT_FFTH	256
#define DFLT_FFTW	1024
#define DFLT_DBMN	(-30.0)
#define DFLT_DBRG	(60.0)

/* static function prototype */
static void usage(void);

/* static data */
       char  *programName;
static int   verb = 0;		/* -v flag */
static int   fftOrder = DFLT_FFTO;
static int   hopSize  = DFLT_FFTH;
static int   winSize  = DFLT_FFTW;
static double dbMin   = DFLT_DBMN;
static double dbRange = DFLT_DBRG;
static char  *inPath  = NULL;
static char  *outPath  = NULL;
static char  defext[16];

/* functions */
void WriteAsciiHex(FILE *file, BYTE *buf, int n)
{	/* write some data out to a file as a line of ascii hexadecimal */
    int i;

    for(i = 0; i<n; ++i)  {
	fprintf(file, "%02x", *buf++);
    }
    fprintf(file, "\n");
}

void DoProcess(char *inpath, char *outpath, int verbose)
{  /* actually do something */
    long len, remain, got, thistime;
    float *sndbuf, *fftbuf, *window;
    BYTE  *bytbuf;
    SOUND *insnd;
    int  chans, i;
    int     winskip, dataskip, dowinlen;
    int  opbins = fftOrder/2+1;
    float   val;
    int     v2;
    int     frames = 0;
#ifdef USEDAT
    long    dim[2];
    int	    tdims;
    const char *key;
    DATFILE  *outfile;
#else /* !USEDAT - write asciihex */
    FILE  *outfile;
#endif /* USEDAT */
#ifdef DODISP 
    WINDAT *wndw = NewWin(programName, 1);
#endif /* DODISP */
#ifdef SGI_FFT
    float *workspace = NULL;
#endif /* SGI_FFT */

    if(fftOrder > winSize)  {
	/* window shorter than fft */
	dataskip = (fftOrder - winSize)/2;
	winskip = 0;
	dowinlen = winSize;
    } else {
	/* window as big as or bigger than fft */
	dataskip = 0;
	winskip  = (winSize - fftOrder)/2;
	dowinlen = fftOrder;
    }

    if( (insnd = sndOpen(NULL, inpath, "r")) == NULL)  {
	fprintf(stderr, "%s: could not read soundfile '%s'\n", 
		programName, inpath);
	return;
    }
#ifdef USEDAT
    if( (outfile = d_open(outpath, "w")) == NULL)  {
#else /* !USEDAT */
    if( (outfile = fopen(outpath, "w")) == NULL)  {
#endif /* USEDAT */
	fprintf(stderr, "%s: could not create output file '%s'\n",
		programName, outpath);
	sndFree(insnd);
	return;
    }
    sndSetUchans(insnd,1);
    sndbuf = TMMALLOC(float, fftOrder, "sndbuf");
    VZero(sndbuf, 1, fftOrder);
    fftbuf = TMMALLOC(float, 2*fftOrder, "fftbuf");
    bytbuf = TMMALLOC(BYTE, opbins, "bytbuf");
    window = TMMALLOC(float, winSize, "window");
    VSetHanning(window, 1, 1.0, winSize);

#ifdef DODISP
    DoDisp(wndw, window, winSize);
#endif /* DODISP */

#ifdef SGI_FFT
    workspace = sfft1dui(fftOrder, workspace);
#endif /* SGI_FFT */

    len = sndGetFrames(insnd);
    remain = len;
#ifdef USEDAT
    /* initilize the DAT file */
    d_settype(outfile, 'b');	/* write data as bytes */
    {	
	char srateStr[32];
	sprintf(srateStr, "%.1f", sndGetSrate(insnd)/hopSize);
	d_setkey(outfile, SAMPLERATE, srateStr);
    }
    dim[0] = (int)(1+len/hopSize);
    dim[1] = opbins;		/* last dim is most rapidly varying */
    {
	char stampStr[128];
	long t;
	time(&t);
	sprintf(stampStr, "Written by tohex at %s", ctime(&t));
	d_setkey(outfile, COMMENT, stampStr);
    }
    d_setdim(outfile, 1, 1, dim, 2);	/* 1 dataset, 1 chan, 2 dimensions */
#else /* !USEDAT */
    /* file header predicts rows and columns */
    fprintf(outfile, "%d %d 8\n%%\n", (int)(1+len/hopSize), opbins);
#endif /* USEDAT */
    fprintf(stderr, "Expecting %d frames of %d cols\n", 
	    (int)(1+len/hopSize), opbins);
    while(remain)  {
	thistime = MIN(remain, hopSize);
	got = sndReadFrames(insnd, sndbuf+fftOrder-hopSize, thistime);
	VCopyZero(sndbuf, 1, fftOrder-hopSize+got, fftbuf, 1, fftOrder);
	/* shift remainder back in sndbuf */
	VCopy(sndbuf+hopSize, 1, sndbuf, 1, fftOrder-hopSize);
	VVMul(window+winskip, 1, fftbuf+dataskip, 1, dowinlen);
#ifdef DODISP
	DoDisp(wndw, fftbuf, fftOrder);
#endif /* DODISP */
#ifdef SGI_FFT
	sfft1du(-1, fftOrder, fftbuf, 1, workspace);
#else /* SGI_FFT */
	fft_packed(fftbuf, fftOrder);
#endif /* SGI_FFT */
	RCVmag((cpxrect *)fftbuf, 1, (cpxpolar *)fftbuf, 1, opbins);
	for(i = 0; i<opbins; ++i)  {
	    val = fftbuf[2*i];
	    if( val <= 0)
		v2 = 0;
	    else
		v2 = 256 * ((4.343*log(val)-dbMin)/dbRange);
	    if(v2 < 0)		v2 = 0;
	    if(v2 > 255)	v2 = 255;
	    bytbuf[i] = v2;
	}
#ifdef USEDAT
	d_writec(outfile, (char *)bytbuf, opbins, 'b');
#else /* !USEDAT */
	WriteAsciiHex(outfile, bytbuf, opbins);
#endif /* USEDAT */
	++frames;
	if(got < thistime )
	    remain = 0;		/* hit end of something */
	else
	    remain -= thistime;
    }
    fprintf(stderr, "%d frames processed, written to %s\n", frames, outpath);
#ifdef USEDAT 
    d_close(outfile);
#else /* !USEDAT */
    fclose(outfile);
#endif /* USEDAT */
    sndFree(insnd);
    return;
}

main(int argc, char **argv)
{
    int 	i = 0;
    int 	err = 0;

    programName = argv[0];
    strcpy(defext, DFLT_EXTN);
/*    strcat(defext, sndfExtn);   */ /* output not a soundfile */

    err = HandleArgs(argc, argv);
    
    if(err || inPath == NULL)
	usage();		/* never returns */

    if(outPath == NULL)  {
	outPath=TMMALLOC(char, strlen(inPath)+strlen(defext), "outPath");
	strcpy(outPath, inPath);
	SubstExtn(inPath, defext, outPath);
    }

    DoProcess(inPath, outPath, verb);
}

int HandleArgs(int argc, char **argv)
{   /* crack command line into static globals */
    int     i = 0, err = 0;

    while(++i<argc && err == 0)  {
	char c, *token, *arg, *nextArg;
	int  argUsed;

	token = argv[i];
	if(*token++ == '-')  {
	    if(*token == '\0')  {  /* - alone -> use stdin */
	    }
	    if(i+1 < argc)
		nextArg = argv[i+1];
	    else
		nextArg = "";
	    argUsed = 0;
	    while(c = *token++)  {
		if(NumericQ(token))	arg = token;
		else			arg = nextArg;
		switch(c)  {
		case 'v':   verb = 1;	break;
		case 'n':
		    fftOrder = atoi(arg);	argUsed = 1;
		    break;
		case 'w':
		    winSize = atoi(arg);	argUsed = 1;
		    break;
		case 'h':
		    hopSize = atoi(arg);	argUsed = 1;
		    break;
		case 'r':
		    dbRange = atof(arg);	argUsed = 1;
		    break;
		case 'm':
		    dbMin = atof(arg);	argUsed = 1;
		    break;
		default:
		    fprintf(stderr,"%s: unrec option %c\n",
			    argv[0], c);
		    err = 1; break;
		}
		if(argUsed)  {
		    if(arg == token)	token = "";   /* no more from token */
		    else		++i;	      /* skip arg we used */
		    arg = ""; argUsed = 0;
		}
	    }
	}
	else{
	    if(inPath == NULL)		inPath = argv[i];
	    else if(outPath == NULL)	outPath = argv[i];
	    else{
		fprintf(stderr,"%s: excess arg %s\n", argv[0], argv[i]);
		err = 1;
	    }
	}
    }
    return err;
}

static void usage(void)		/* print syntax & exit */
{
    fprintf(stderr,
    "usage: %s [-n n][-w n][-h n][-b n][-r f][-m f] inSnd [outHex]\n", programName);
    fprintf(stderr,"where (defaults in brackets)\n");
    fprintf(stderr," -n n  number of points in FFT        (%d)\n",DFLT_FFTO);
    fprintf(stderr," -w n  window point size (<= fftsize) (%d)\n",DFLT_FFTW);
    fprintf(stderr," -h n  hop increment points           (%d)\n",DFLT_FFTH);
    fprintf(stderr," -r f  black-to-white range (dB)      (%.1f)\n",DFLT_DBRG);
    fprintf(stderr," -m f  min mag (black pt) (dB)        (%.1f)\n",DFLT_DBMN);
    fprintf(stderr," inSnd    input sound               ('-' uses stdin)\n");
    fprintf(stderr," outHex   output hex-ascii data     (instem%s)\n", DFLT_EXTN);
    exit(1);
}

